# Scola version 1.1.0
# ScolaBacktrackStack.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) BacktrackInstruction
# 3) BacktrackBlocks
# 4) BacktrackStacks

# 1) Imported modules
# -------------------

import sys
import os

from ScolaConstants import *
from ScolaModels import *
from ScolaHistories import *
from ScolaProcesses import *

# 2) BacktrackInstructions
# ------------------------

class ScolaBacktrackInstruction:
  NEW_PROCESS = "NEW_PROCESS"
  ACTIVATE_PROCESS = "ACTIVATE_PROCESS"
  DEACTIVATE_PROCESS = "DEACTIVATE_PROCESS"
  MOVE_PROCESS = "MOVE_PROCESS"
  ENQUEUE_PROCESS = "ENQUEUE_PROCESS"
  DEQUEUE_PROCESS = "DEQUEUE_PROCESS"
  SET_PORT = "SET_PORT"
  NEW_ELEMENT = "NEW_ELEMENT"
  DELETE_ELEMENT = "DELETE_ELEMENT"
  MOVE_ELEMENT = "MOVE_ELEMENT"
  SEND_ELEMENT = "SEND_ELEMENT"
  RECEIVE_ELEMENT = "RECEIVE_ELEMENT"
  CLONE_ELEMENT = "CLONE_ELEMENT"

  def __init__(self, type, size):
    self._type = type
    self._size = size
    self._data = [None for _ in range(0, size)]

  def GetType(self):
    return self._type
  def GetSize(self):
    return self._size
  def GetDatum(self, index):
    if index<0 or index>=self.GetSize():
      return None
    return self._data[index]
  def SetDatum(self, index, datum):
    if index<0 or index>=self.GetSize():
      return
    self._data[index] = datum

  @staticmethod
  def NewNewProcess(process):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.NEW_PROCESS, 1)
    self.SetDatum(0, process)
    return self

  @staticmethod
  def NewActivateProcess(process):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.ACTIVATE_PROCESS, 1)
    self.SetDatum(0, process)
    return self

  @staticmethod
  def NewDeactivateProcess(process, location):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.DEACTIVATE_PROCESS, 2)
    self.SetDatum(0, process)
    self.SetDatum(1, location)
    return self

  @staticmethod
  def NewMoveProcess(process, element):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.MOVE_PROCESS, 2)
    self.SetDatum(0, process)
    self.SetDatum(1, element)
    return self

  @staticmethod
  def NewEnqueueProcess(process, element):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.ENQUEUE_PROCESS, 2)
    self.SetDatum(0, process)
    self.SetDatum(1, element)
    return self

  @staticmethod
  def NewDequeueProcess(process, element, index):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.DEQUEUE_PROCESS, 3)
    self.SetDatum(0, process)
    self.SetDatum(1, element)
    self.SetDatum(2, index)
    return self

  @staticmethod
  def NewSetPort(port, value):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.SET_PORT, 2)
    self.SetDatum(0, port)
    self.SetDatum(1, value)
    return self

  @staticmethod
  def NewNewElement(element):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.NEW_ELEMENT, 1)
    self.SetDatum(0, element)
    return self

  @staticmethod
  def NewDeleteElement(container, element, index):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.DELETE_ELEMENT, 3)
    self.SetDatum(0, container)
    self.SetDatum(1, element)
    self.SetDatum(2, index)
    return self

  @staticmethod
  def NewMoveElement(container, element, index, identifier):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.MOVE_ELEMENT, 4)
    self.SetDatum(0, container)
    self.SetDatum(1, element)
    self.SetDatum(2, identifier)
    self.SetDatum(3, index)
    return self

  @staticmethod
  def NewSendElement(sourceBlock, element, index, receiver, key):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.SEND_ELEMENT, 5)
    self.SetDatum(0, sourceBlock)
    self.SetDatum(1, element)
    self.SetDatum(2, index)
    self.SetDatum(3, receiver)
    self.SetDatum(4, key)
    return self

  @staticmethod
  def NewReceiveElement(element, identifier, receiver, key):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.RECEIVE_ELEMENT, 4)
    self.SetDatum(0, element)
    self.SetDatum(1, identifier)
    self.SetDatum(2, receiver)
    self.SetDatum(3, key)
    return self

  @staticmethod
  def NewCloneElement(element):
    self = ScolaBacktrackInstruction(ScolaBacktrackInstruction.CLONE_ELEMENT, 1)
    self.SetDatum(0, element)
    return self

# 3) BacktrackBlocks
# ------------------

class ScolaBacktrackBlock:
  def __init__(self, command):
    self._command = command
    self._instructions = []

  def GetCommand(self):
    return self._command

  def IsEmpty(self):
    return len(self._instructions)==0
  def PushInstruction(self, instruction):
    self._instructions.append(instruction)
  def PopInstruction(self):
    return self._instructions.pop()

# 4) BacktrackStacks
# ------------------

class ScolaBacktrackStack:
  def __init__(self):
    self._blocks = []

  def Reset(self):
    self._blocks = []

  def IsEmpty(self):
    return len(self._blocks)==0
  def GetSize(self):
     return len(self._blocks)
  def GetTopBlock(self):
    if self.IsEmpty():
      return None
    return self._blocks[-1]
  def GetBlock(self, index):
    if index<0 or index>=self.GetSize():
      return None
    return self._blocks[index]
  def PushBlock(self, command):
    block = ScolaBacktrackBlock(command)
    self._blocks.append(block)
    return block
  def PopBlock(self):
    return self._blocks.pop()

  def NewNewProcess(self, process):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewNewProcess(process)
    block.PushInstruction(instruction)

  def NewActivateProcess(self, process):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewActivateProcess(process)
    block.PushInstruction(instruction)

  def NewDeactivateProcess(self, process, location):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewDeactivateProcess(process, location)
    block.PushInstruction(instruction)

  def NewMoveProcess(self, process, element):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewMoveProcess(process, element)
    block.PushInstruction(instruction)

  def NewEnqueueProcess(self, process, element):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewEnqueueProcess(process, element)
    block.PushInstruction(instruction)

  def NewDequeueProcess(self, process, element, index):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewDequeueProcess(process, element, index)
    block.PushInstruction(instruction)

  def NewSetPort(self, port, value):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewSetPort(port, value)
    block.PushInstruction(instruction)

  def NewNewElement(self, element):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewNewElement(element)
    block.PushInstruction(instruction)

  def NewDeleteElement(self, container, element, index):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewDeleteElement(container, element, index)
    block.PushInstruction(instruction)

  def NewMoveElement(self, container, element, index, identifier):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewMoveElement(container, element, index, identifier)
    block.PushInstruction(instruction)

  def NewSendElement(self, sourceBlock, element, index, receiver, key):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewSendElement(sourceBlock, element, index, receiver, key)
    block.PushInstruction(instruction)

  def NewReceiveElement(self, element, identifier, receiver, key):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewReceiveElement(element, identifier, receiver, key)
    block.PushInstruction(instruction)

  def NewCloneElement(self, element):
    if self.IsEmpty():
      return
    block = self.GetTopBlock()
    instruction = ScolaBacktrackInstruction.NewCloneElement(element)
    block.PushInstruction(instruction)
