# Scola version 1.1.0
# ScolaExpressionEvaluator.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) ExpressionEvaluator

# 1) Imported modules
# -------------------

import sys
import os

from ScolaModels import *

# 2) ExpressionEvaluator
# ---------------------

# 2.1) Creator
# ------------

class ScolaExpressionEvaluator:
  def __init__(self):
    self._model = None
    self._systemPath = None
    self._quoteLevel = 0

  def GetModel(self):
    return self._model
  def SetModel(self, model):
    self._model = model

  def GetSystemPath(self):
    return self._systemPath
  def SetSystemPath(self, path):
    self._systemPath = path

# 2.2) Evaluation as constants
# ----------------------------

  def EvalExpression(self, expression):
    self._quoteLevel = 0
    return self._EvalExpression(expression)

  def _EvalExpression(self, expression):

    if expression.GetType()==ScolaExpression.REFERENCE:
      if self._quoteLevel>1:
        return ScolaConstant.NewNone()
      if self._quoteLevel==1:
        value = expression.GetValue()
        return ScolaConstant.Copy(value)
      value = expression.GetValue()
      level = self._quoteLevel
      while level<=0:
        level += 1
        if not value.IsSymbol():
          return ScolaConstant.NewNone()
        reference = value.GetValueAsSymbol()
        path = ScolaPath.FromString(reference)
        path = ScolaPath.Merge(self._systemPath, path)
        element = self._model.AccessNamedElement(path)
        if element==None or element.GetType()!=ScolaModel.PORT:
          return ScolaConstant.NewNone()
        value = element.GetValue()
      return ScolaConstant.Copy(value)

    if expression.GetType()==ScolaExpression.SYMBOL:
      value = expression.GetValue()
      return ScolaConstant.Copy(value)

    if expression.GetType()==ScolaExpression.NONE:
      return ScolaConstant.NewNone()

    if expression.GetType()==ScolaExpression.TRUE:
      return ScolaConstant.NewTrue()

    if expression.GetType()==ScolaExpression.FALSE:
      return ScolaConstant.NewFalse()

    if expression.GetType()==ScolaExpression.AND:
      value = ScolaConstant.NewTrue()
      childValue = ScolaExpression.NewNone()
      for child in expression.GetChildren():
        childValue = self._EvalExpression(child)
        value = ScolaConstant.And(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.OR:
      value = ScolaConstant.NewFalse()
      childValue = ScolaExpression.NewNone()
      for child in expression.GetChildren():
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Or(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.NOT:
      child = expression.GetChild(0)
      childValue = self._EvalExpression(child)
      return ScolaConstant.Not(childValue)

    if expression.GetType()==ScolaExpression.EQ:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Eq(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.DF:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Df(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.LT:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Lt(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.GT:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Gt(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.LEQ:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Leq(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.GEQ:
      child1 = expression.GetChild(0)
      childValue1 = self._EvalExpression(child1)
      child2 = expression.GetChild(1)
      childValue2 = self._EvalExpression(child2)
      return ScolaConstant.Geq(childValue1, childValue2)

    if expression.GetType()==ScolaExpression.INTEGER:
      return ScolaConstant.Copy(expression.GetValue())

    if expression.GetType()==ScolaExpression.REAL:
      return ScolaConstant.Copy(expression.GetValue())

    if expression.GetType()==ScolaExpression.OPP:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Opp(childValue)
      return value

    if expression.GetType()==ScolaExpression.INV:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Inv(childValue)
      return value

    if expression.GetType()==ScolaExpression.ADD:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Add(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.SUB:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Sub(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.MUL:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Mul(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.DIV:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Div(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.MIN:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Min(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.MAX:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsNumber():
        return ScolaConstant.NewNone()
      for i in range(1, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        value = ScolaConstant.Max(value, childValue)
      return value

    if expression.GetType()==ScolaExpression.ABS:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Abs(childValue)
      return value

    if expression.GetType()==ScolaExpression.POW:
      child1 = expression.GetChild(0)
      child2 = expression.GetChild(1)
      if child1==None or child2==None:
        return ScolaConstant.NewNone()
      childValue1 = self._EvalExpression(child1)
      childValue2 = self._EvalExpression(child2)
      value = ScolaConstant.Pow(childValue1, childValue2)
      return value

    if expression.GetType()==ScolaExpression.EXP:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Exp(childValue)
      return value

    if expression.GetType()==ScolaExpression.LOG:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Log(childValue)
      return value

    if expression.GetType()==ScolaExpression.SQRT:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Sqrt(childValue)
      return value

    if expression.GetType()==ScolaExpression.MOD:
      child1 = expression.GetChild(0)
      child2 = expression.GetChild(1)
      if child1==None or child2==None:
        return ScolaConstant.NewNone()
      childValue1 = self._EvalExpression(child1)
      childValue2 = self._EvalExpression(child2)
      value = ScolaConstant.Mod(childValue1, childValue2)
      return value

    if expression.GetType()==ScolaExpression.CEIL:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Ceil(childValue)
      return value

    if expression.GetType()==ScolaExpression.FLOOR:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      childValue = self._EvalExpression(child)
      value = ScolaConstant.Floor(childValue)
      return value

    if expression.GetType()==ScolaExpression.COUNT:
      count = 0
      for i in range(0, expression.GetNumberOfChildren()):
        child = expression.GetChild(i)
        childValue = self._EvalExpression(child)
        if childValue.IsTrue():
          count += 1
        elif not childValue.IsFalse():
          return ScolaConstant.NewNone()
      return ScolaConstant.NewInteger(count)

    if expression.GetType()==ScolaExpression.STRING:
      return ScolaConstant.Copy(expression.GetValue())

    if expression.GetType()==ScolaExpression.APPEND:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if value.IsSymbol():
        for i in range(1, expression.GetNumberOfChildren()):
          child = expression.GetChild(i)
          childValue = self._EvalExpression(child)
          if not childValue.IsSymbol():
            return ScolaConstant.NewNone()
          symbol = value.GetValueAsSymbol() + ScolaPath.COMPOSITION_SEPARATOR + childValue.GetValueAsSymbol()
          value = ScolaConstant.NewSymbol(symbol)
        return value
      if value.IsString():
        for i in range(1, expression.GetNumberOfChildren()):
          child = expression.GetChild(i)
          childValue = self._EvalExpression(child)
          value = ScolaConstant.Append(value, childValue)
        return value
      return ScolaConstant.NewNone()

    if expression.GetType()==ScolaExpression.IF:
      condition = expression.GetChild(0)
      value = self._EvalExpression(condition)
      if value.IsTrue():
         return self._EvalExpression(expression.GetChild(1))
      if value.IsFalse():
        return self._EvalExpression(expression.GetChild(2))
      return ScolaConstant.NewNone()

    if expression.GetType()==ScolaExpression.CAST_INTEGER:
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      return ScolaConstant.ToInteger(value)

    if expression.GetType()==ScolaExpression.CAST_REAL:
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      return ScolaConstant.ToReal(value)

    if expression.GetType()==ScolaExpression.CAST_STRING:
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      return ScolaConstant.ToString(value)

    if expression.GetType()==ScolaExpression.CAST_SYMBOL:
      if expression.GetNumberOfChildren()==0:
        return ScolaConstant.NewSymbol("")
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      if value.IsSymbol():
        return value
      if value.IsString():
        string = value.GetValueAsString()
        if ScolaPath.IsValidPath(string):
          return ScolaConstant.NewSymbol(string)
      return ScolaConstant.NewNone()

    if expression.GetType()==ScolaExpression.IDENTIFIER:
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      if not value.IsSymbol():
        return ScolaConstant.NewNone()
      symbol = value.GetValueAsSymbol()
      if not ScolaPath.IsValidPath(symbol):
        return ScolaConstant.NewNone()
      path = ScolaPath.FromString(symbol)
      identifier = path.GetLastIdentifier()
      return ScolaConstant.NewSymbol(identifier)

    if expression.GetType()==ScolaExpression.OWNER:
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      if not value.IsSymbol():
        return ScolaConstant.NewNone()
      symbol = value.GetValueAsSymbol()
      if not ScolaPath.IsValidPath(symbol):
        return ScolaConstant.NewNone()
      path = ScolaPath.FromString(symbol)
      path.RemoveLastIdentifier()
      return ScolaConstant.NewSymbol(path.ToString())

    if expression.GetType()==ScolaExpression.IS_BLOCK:
      child = expression.GetChild(0)
      path = self._EvalExpressionAsPath(child)
      if path==None:
        return ScolaConstant.NewNone()
      element = self._model.AccessNamedElement(path)
      if element==None or element.GetType()!=ScolaModel.BLOCK:
        return ScolaConstant.NewFalse()
      return ScolaConstant.NewTrue()

    if expression.GetType()==ScolaExpression.IS_PORT:
      child = expression.GetChild(0)
      path = self._EvalExpressionAsPath(child)
      if path==None:
        return ScolaConstant.NewNone()
      element = self._model.AccessNamedElement(path)
      if element==None or element.GetType()!=ScolaModel.PORT:
        return ScolaConstant.NewFalse()
      return ScolaConstant.NewTrue()

    if expression.GetType()==ScolaExpression.IS_ASSERTION:
      child = expression.GetChild(0)
      path = self._EvalExpressionAsPath(child)
      if path==None:
        return ScolaConstant.NewNone()
      element = self._model.AccessNamedElement(path)
      if element==None or element.GetType()!=ScolaModel.ASSERTION:
        return ScolaConstant.NewFalse()
      return ScolaConstant.NewTrue()

    if expression.GetType()==ScolaExpression.SIZE:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsSymbol():
        return ScolaConstant.NewNone()
      path = ScolaPath.FromString(value.GetValueAsSymbol())
      if path==None:
        return ScolaConstant.NewNone()
      absolutePath = ScolaPath.Merge(self._systemPath, path)
      container = self._model.AccessNamedElement(absolutePath)
      if container==None or container.GetType()!=ScolaModel.BLOCK:
        return ScolaConstant.NewNone()
      size = container.GetNumberOfElements()
      return ScolaConstant.NewInteger(size)

    if expression.GetType()==ScolaExpression.ELEMENT:
      child = expression.GetChild(0)
      if child==None:
        return ScolaConstant.NewNone()
      value = self._EvalExpression(child)
      if not value.IsSymbol():
        return ScolaConstant.NewNone()
      path = ScolaPath.FromString(value.GetValueAsSymbol())
      if path==None:
        return ScolaConstant.NewNone()
      child = expression.GetChild(1)
      index = self._EvalExpression(child)
      if not index.IsInteger():
        return ScolaConstant.NewNone()
      absolutePath = ScolaPath.Merge(self._systemPath, path)
      container = self._model.AccessNamedElement(absolutePath)
      if container==None or container.GetType()!=ScolaModel.BLOCK:
        return ScolaConstant.NewNone()
      element = container.GetElement(index.GetValueAsInteger())
      if element==None:
        return ScolaConstant.NewNone()
      path.AppendIdentifier(element.GetIdentifier())
      return ScolaConstant.NewSymbol(path.ToString())

    if expression.GetType()==ScolaExpression.QUOTE:
      oldQuoted = self._quoteLevel
      self._quoteLevel += 1
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      self._quoteLevel -= 1
      return value

    if expression.GetType()==ScolaExpression.EVAL:
      self._quoteLevel -= 1
      child = expression.GetChild(0)
      value = self._EvalExpression(child)
      self._quoteLevel += 1
      return value

    # Should not happen
    print("horror")
    return ScolaConstant.NewNone()

# 2.2) Evaluation as paths
# ------------------------

  def EvalExpressionAsPath(self, expression):
    oldQuoted = self._quoteLevel
    self._quoteLevel = 0
    path = self._EvalExpressionAsPath(expression)
    self._quoteLevel = oldQuoted
    return path

  def _EvalExpressionAsPath(self, expression):
    self._quoteLevel += 1
    value = self._EvalExpression(expression)
    self._quoteLevel -= 1
    if not value.IsSymbol():
      return None
    path = ScolaPath.FromString(value.GetValueAsSymbol())
    return ScolaPath.Merge(self._systemPath, path)
