# Scola version 1.1.0
# ScolaShell.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Shell
# 3) Commands

# 1) Imported modules
# -------------------

import sys
import os
import re
import cmd
# On Linux and Mac OS, decomment the following line
# import readline

from ScolaWalker import *

# 2) Shell
# --------

class ScolaShell(cmd.Cmd):
  intro = 'Welcome to the Scola shell.   Type help or ? to list commands.\n'
  prompt = '(Scola) '
  # On Linux and Mac OS, set the variable use_rawinput to True 
  use_rawinput = False
  walker = ScolaWalker()

  # ----- OS commands -----
  def do_ls(self, arg):
    'ls [<dirname>]\nList files in the given directory (by default the current working directory\n'
    self.ListFiles(arg)

  # ----- Scola commands -----
  def do_load(self, arg):
    """load <filename>
Load a Scola model from a file"""
    self.LoadModel(arg)
  def do_reset(self, arg):
    """reset
Reset the model"""
    self.ResetModel()
  def do_compile(self, arg):
    """compile
Compile the model so to make it possible to execute it"""
    self.CompileModel()
  def do_exit(self, arg):
    """exit
Stop the Scola interpreted, close the Scola window and exit"""
    print('Thank you for using Scola')
    sys.exit()
    return True

  # ----- Scola simulations -----
  def do_start(self, arg):
    """start
Start a simulation"""
    self.StartSimulation()

  def do_stop(self, arg):
    """stop
Stop the current simulation"""
    self.StopSimulation()

  def do_new(self, arg):
    """new <state>*
Create a new process at each of the given (initial) states
new -a | --all creates a new process at each initial state of the model"""
    self.NewProcess(arg)

  def do_kill(self, arg):
    """kill <number>*
Kill the given processes"""
    self.KillProcess(arg)

  def do_next(self, arg):
    """next <number> [<choice-branch>]
Move the given processes forward, if possible.
If the process is on a choice gateway the branch to follow must be given as second argument"""
    self.MoveProcessForward(arg)

  def do_back(self, arg):
    """back [<option>]
Options:
- no option given: undo the last simulation step
- -a | --all: undo all simulation steps performed so far
- <integer>: undo the given number of simulation steps"""
    self.Backtrack(arg)

  def do_findall(self, arg):
    """findall [<filename>]
Find (and print) all histories starting from the current state"""
    self.FindAll(arg)

  def do_findone(self, arg):
    """findone [<filename>]
Find (and print) the first history starting from the current state"""
    self.FindOne(arg)

  def do_findnext(self, arg):
    """findnext [<filename>]
Find (and print) the next history"""
    self.FindNext(arg)

  def do_set(self, arg):
    """set <option>
Options:
- max-num: maximum number of histories a walk is looking for.
- max-len: maximim length of an history a walk is looking for.
- max-stp: maximum number of steps of a walk."""
    self.Set(arg)

  def do_print(self, arg):
    """print [<options>] <target>
Targets:
- model: print the model.
- blocks: print information on all top-level blocks.
- block <path>*: print information on the given blocks.
- processes: print information on all active processes.
- process <number>*: print information on the given processes.
- nexts: print next locations of all active processes.
- next <number>*: print next locations of the given processes.
- history: print actions of the current history.
- max-num: maximum number of histories a walk is looking for.
- max-len: maximim length of an history a walk is looking for.
- max-stp: maximum number of steps of a walk.
Options:
- -o | --overwrite <filename>: prints in the given file in overwrite mode
- -a | --append <filename>: prints in the given file is append mode"""
    self.Print(arg)

  def do_play(self, arg):
    """play <filename>*:
Play the given Scola scripts"""
    self.PlayScript(arg)

  # ----- Management functions -----
  def precmd(self, line):
    line = re.sub(r'[ \t]*#.*', r'', line)
    return line
  def emptyline(self):
    pass

  def complete_new(self, text, line, begidx, endidx):
    return [state for state in self.walker.GetInitialStatePaths() if state.startswith(text)]

# 3) Commands
# -----------

  def ListFiles(self, line):
    args = line.split()
    if len(args)==0:
      filename = ''
      path = './'
    else:
      path = args[0]
      if '/' in path:
        i = path.rfind('/')
        filename = path[i+1:]
        path = path[:i]
      else:
        filename = path
        path = './'
    ls = os.listdir(path)
    # ls = ls[:] # for overwrite in annotate.
    # dircache.annotate(path, ls)
    if filename!='':
      ls = [f for f in ls if f.startswith(filename)]
    for f in ls:
      print(f)

  def LoadModel(self, inputFileName):
    inputFileName = inputFileName.strip('"')
    try:
      inputFile = open(inputFileName, "r")
    except:
      self.RaiseError('Unable to open file "%s"\n' % inputFileName)
      return
    numberOfErrors = self.walker.ParseFile(inputFileName, inputFile)
    inputFile.close()
    if numberOfErrors==0:
      print("Loading file %s completed\n" % inputFileName)
    else:
      print("Loading file %s failed: %d errors detected\n" % (inputFileName, numberOfErrors))

  def ResetModel(self):
    self.walker.ResetModel()
    print("Model reseted\n")

  def CompileModel(self):
    numberOfErrors = self.walker.CompileModel()
    print("Model checked: %d errors\n" % numberOfErrors)

  def StartSimulation(self):
    if not self.walker.IsModelCompiled():
      self.CompileModel()
      if not self.walker.IsModelCompiled():
        return
    self.walker.StartSimulation()
    print("Simulation started\n")

  def StopSimulation(self):
    self.walker.StopSimulation()

  def Backtrack(self, line):
    args = line.split()
    if len(args)==0:
      self.walker.Backtrack()
      return
    if args[0]=="-a" or args[0]=="--all":
      self.walker.BacktrackAll()
      return
    if not self.IsInteger(args[0]):
      self.RaiseError("An integer was expected")
      return
    numberOfSteps = int(args[0])
    for _ in range(0, numberOfSteps):
      self.walker.Backtrack()

  def NewProcess(self, line):
    args = line.split()
    if len(args)==0:
      self.RaiseError("Wrong number of arguments")
      return
    if args[0]=="-a" or args[0]=="--all":
      states = self.walker.GetPathsOfInitialStates()
    else:
      states = args
    for state in states:
      number = self.walker.NewProcess(state)
      if number!=-1:
        print("Process %d created at %s\n" % (number, state))

  def KillProcess(self, line):
    args = line.split()
    numbers = [int(arg) for arg in args if self.IsInteger(arg)]
    self.walker.DeactivateProcesses(numbers)

  def MoveProcessForward(self, line):
    args = line.split()
    if len(args)==0:
      self.RaiseError("Wrong number of arguments")
      return
    if not self.IsInteger(args[0]):
      self.RaiseError("A process number was expected")
      return
    number = int(args[0])
    if len(args)==1:
      branch = None
    else:
      branch = args[1]
    self.walker.MoveProcessForward(number, branch)

  def FindAll(self, line):
    self.walker.FindAll()

  def FindOne(self, line):
    self.walker.FindOne()

  def FindNext(self, line):
    self.walker.FindNext()

  def Set(self, line):
    args = line.split()
    i = 0
    if i==len(args):
      self.RaiseError("Wrong number of arguments")
      return
    if args[i]=="max-num":
      i += 1
      if not self.IsInteger(args[i]):
        self.RaiseError("An integer was expected")
        return
      self.walker.SetMaximumNumberOfHistories(int(args[i]))
    elif args[i]=="max-len":
      i += 1
      if not self.IsInteger(args[i]):
        self.RaiseError("An integer was expected")
        return
      self.walker.SetMaximumSizeOfHistories(int(args[i]))
    elif args[i]=="max-stp":
      i += 1
      if not self.IsInteger(args[i]):
        self.RaiseError("An integer was expected")
        return
      self.walker.SetMaximumNumberOfStepsOfWalks(int(args[i]))
    else:
      self.RaiseError("%s is not an option of the command set" % args[i])

  def Print(self, line):
    args = line.split()
    i = 0
    if i==len(args):
      self.RaiseError("Wrong number of arguments")
      return
    previousOutputFile = self.walker.GetOutputFile()
    (outputFile, j) = self.GetOutputFileOption(args, i)
    if outputFile==None and j!=i:
      return
    i = j
    if i==len(args):
      self.RaiseError("Wrong number of arguments")
      return
    if outputFile!=None:
      self.walker.SetOutputFile(outputFile)
    if args[i]=="model":
      self.walker.PrintModel()
    elif args[i]=="compiled-model":
      self.walker.PrintCompiledModel()
    elif args[i]=="blocks":
      self.walker.PrintBlocks(self.walker.GetIdentifiersOfTopLevelBlocks())
    elif args[i]=="block":
      self.walker.PrintBlocks(args[i:])
    elif args[i]=="processes":
      self.walker.PrintProcesses(self.walker.GetNumbersOfActiveProcesses())
    elif args[i]=="process":
      numbers = [int(arg) for arg in args[i+1:] if self.IsInteger(arg)]
      self.walker.PrintProcesses(numbers)
    elif args[i]=="nexts":
      self.walker.PrintNextLocationsOfProcesses(self.walker.GetNumbersOfActiveProcesses())
    elif args[i]=="next":
      numbers = [int(arg) for arg in args[i+1:] if self.IsInteger(arg)]
      self.walker.PrintNextLocationsOfProcesses(numbers)
    elif args[i]=="history":
      self.walker.PrintHistory()
    elif args[i]=="max-num":
      self.walker.Println("set max-num " + str(self.walker.GetMaximumNumberOfHistories()))
    elif args[i]=="max-len":
      self.walker.Println("set max-len " + str(self.walker.GetMaximumSizeOfHistories()))
    elif args[i]=="max-stp":
      self.walker.Println("set max-stp " + str(self.walker.GetMaximumNumberOfStepsOfWalks()))
    else:
      self.RaiseError("%s is not an option of the command print" % args[i])
    if outputFile!=None:
      self.walker.SetOutputFile(previousOutputFile)
      outputFile.flush()
      outputFile.close()

  def PlayScript(self, line):
    args = line.split()
    for scriptFileName in args:
      try:
        scriptFile = open(scriptFileName, "r")
      except:
        self.RaiseError('Unable to open file "%s"\n' % scriptFileName)
        continue
      self.cmdqueue.extend(scriptFile.read().splitlines())

  def IsInteger(self, arg):
    try:
      int(arg)
    except:
      return False
    return True

  def GetOutputFileOption(self, args, j):
    i = j
    if args[i]=="-o":
      i += 1
      if i==len(args):
        self.RaiseError("Wrong number of arguments")
        return (None, i)
      fileName = args[i]
      i += 1
      try:
        outputFile = open(fileName, "w")
        return (outputFile, i)
      except:
        self.RaiseError('Unable to open file "%s"\n' % fileName)
        return (None, i)
    if args[i]=="-a":
      i += 1
      if i==len(args):
        self.RaiseError("Wrong number of arguments")
        return (None, i)
      fileName = args[i]
      i += 1
      try:
        outputFile = open(fileName, "a")
        return (outputFile, i)
      except:
        self.RaiseError('Unable to open file "%s"\n' % fileName)
        return (None, i)
    return (None, i)
    

  def RaiseError(self, message):
    self.walker.RaiseError(message)
