# Scola version 1.1.0
# ScolaWalker.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Walker
# 3) WalkStack

import sys
import os

from ScolaStepper import ScolaStepper

# 2) Walker
# ---------

# 2.1) Creator
# ------------

class ScolaWalker:
  FORTH = "FORTH"
  BACK = "BACK"

  CONTINUING = "CONTINUING"
  INTERRUPTED = "INTERRUPTED"
  TERMINATED = "TERMINATED"

  traceFile = None

  def __init__(self):
    self._stepper = ScolaStepper()
    self._walkStack = ScolaWalkStack()
    self._status = ScolaWalker.FORTH
    self._maximumNumberOfHistories = 1000
    self._maximumSizeOfHistories = 10000
    self._maximumNumberOfStepsOfWalks = 1000000
    self._numberOfHistories = 0
    self._numberOfSteps = 0
    self._walkStarted = False

  def GetMaximumNumberOfHistories(self):
    return self._maximumNumberOfHistories
  def SetMaximumNumberOfHistories(self, value):
    self._maximumNumberOfHistories = value

  def GetMaximumSizeOfHistories(self):
    return self._maximumSizeOfHistories
  def SetMaximumSizeOfHistories(self, value):
    self._maximumSizeOfHistories = value

  def GetMaximumNumberOfStepsOfWalks(self):
    return self._maximumNumberOfStepsOfWalks
  def SetMaximumNumberOfStepsOfWalks(self, value):
    self._maximumNumberOfStepsOfWalks = value

# 2.2) Stepper interface
# ----------------------

  # 2.2.1) Model
  #-------------

  def ParseFile(self, inputFileName, inputFile):
    numberOfErrors = self._stepper.ParseFile(inputFileName, inputFile)
    return numberOfErrors

  def ResetModel(self):
    self._stepper.ResetModel()

  def CompileModel(self):
    numberOfErrors = self._stepper.CompileModel()
    return numberOfErrors
  def IsModelCompiled(self):
    return self._stepper.IsModelCompiled()

  def GetPathsOfInitialStates(self):
    return self._stepper.GetPathsOfInitialStates()

  def GetIdentifiersOfTopLevelBlocks(self):
    return self._stepper.GetIdentifiersOfTopLevelBlocks()

  # 2.2.2) Simulation
  #------------------

  def StartSimulation(self):
    self._walkStarted = False
    ok = self._stepper.StartSimulation()
    return ok

  def StopSimulation(self):
    self._walkStarted = False
    ok = self._stepper.StopSimulation()
    return ok

  def NewProcess(self, state):
    processNumber = self._stepper.NewProcess(state)
    return processNumber

  def DeactivateProcesses(self, processNumbers):
    numberOfDeactivatedProcesses = self._stepper.DeactivateProcesses(processNumbers)
    return numberOfDeactivatedProcesses

  def MoveProcessForward(self, processNumber, branchName):
    ok = self._stepper.MoveProcessForward(processNumber, branchName)
    return ok

  def GetBacktrackLimit(self):
    return self._stepper.GetBacktrackLimit()
  def SetBacktrackLimit(self, limit):
    self._stepper.SetBacktrackLimit(limit)
  def SetBacktrackLimitHere(self):
    self._stepper.SetBacktrackLimitHere()

  def Backtrack(self):
    ok = self._stepper.Backtrack()
    return ok

  def BacktrackAll(self):
    ok = self._stepper.BacktrackAll()
    return ok

  def GetNumbersOfActiveProcesses(self):
    return self._stepper.GetNumbersOfActiveProcesses()

  def IsProcessLocatedOnATerminalState(self, processNumber):
    return self._stepper.IsProcessLocatedOnATerminalState(processNumber)

  def IsProcessLocatedOnAChoiceGateway(self, processNumber):
    return self._stepper.IsProcessLocatedOnAChoiceGateway(processNumber)

  def GetBranchesOfLocationOfProcess(self, processNumber):
    return self._stepper.GetBranchesOfLocationOfProcess(processNumber)

  # 2.2.3) Printing
  #----------------

  def GetOutputFile(self):
    return self._stepper.GetOutputFile()
  def SetOutputFile(self, file):
    self._stepper.SetOutputFile(file)

  def GetErrorFile(self):
    return self._stepper.GetErrorFile()
  def SetErrorFile(self, file):
    self._stepper.SetErrorFile(file)

  def IsVerbose(self):
    return self._stepper.IsVerbose()
  def SetVerbose(self, value):
    self._stepper.SetVerbose(value)

  def PrintModel(self):
    self._stepper.PrintModel()

  def PrintCompiledModel(self):
    self._stepper.PrintCompiledModel()

  def PrintBlocks(self, paths):
    self._stepper.PrintBlocks(paths)

  def PrintProcesses(self, processNumbers):
    self._stepper.PrintProcesses(processNumbers)

  def PrintNextLocationsOfProcesses(self, processNumbers):
    self._stepper.PrintNextLocationsOfProcesses(processNumbers)

  def PrintHistory(self):
    self._stepper.PrintHistory()

  def Print(self, *texts):
    self._stepper.Print(*texts)
  def Println(self, *texts):
    self._stepper.Println(*texts)

  def RaiseError(self, message):
    self._stepper.RaiseError(message)

# 2.3) Walks
# ----------

  def FindAll(self):
    self._walkStarted = False
    verbose = self.IsVerbose()
    self.SetVerbose(False)
    self._walkStack.Reset()
    self._numberOfHistories = 0
    self._numberOfSteps = 0
    choiceQueue = self._walkStack.PushChoiceQueue()
    self._EnqueueChoices(choiceQueue)
    if choiceQueue.IsEmpty():
      self._stepper.PrintHistory()
      self._status = ScolaWalker.BACK
      return ScolaWalker.TERMINATED
    else:
      self._status = ScolaWalker.FORTH
    while True:
      walkStatus = self._WalkNext()
      if walkStatus!=ScolaWalker.CONTINUING:
        break
    self.SetVerbose(verbose)
    return walkStatus

  def FindOne(self):
    self._walkStarted = True
    verbose = self.IsVerbose()
    self.SetVerbose(False)
    self._walkStack.Reset()
    self._numberOfHistories = 0
    self._numberOfSteps = 0
    choiceQueue = self._walkStack.PushChoiceQueue()
    self._EnqueueChoices(choiceQueue)
    if choiceQueue.IsEmpty():
      # self._stepper.PrintHistory()
      self._status = ScolaWalker.BACK
      self._walkStarted = True
      return ScolaWalker.TERMINATED
    else:
      self._status = ScolaWalker.FORTH
    maximumNumberOfHistories = self._maximumNumberOfHistories
    self._maximumNumberOfHistories = 1
    while True:
      walkStatus = self._WalkNext()
      if walkStatus!=ScolaWalker.CONTINUING:
        break
    self.SetVerbose(verbose)
    self._maximumNumberOfHistories = maximumNumberOfHistories
    if walkStatus==ScolaWalker.TERMINATED:
      self._walkStarted = True
    return walkStatus

  def FindNext(self):
    if not self._walkStarted:
      self.RaiseError("Find walk not started or terminated")
      return
    verbose = self.IsVerbose()
    self.SetVerbose(False)
    maximumNumberOfHistories = self._maximumNumberOfHistories
    self._maximumNumberOfHistories = self._numberOfHistories+1
    while True:
      walkStatus = self._WalkNext()
      if walkStatus!=ScolaWalker.CONTINUING:
        break
    self.SetVerbose(verbose)
    self._maximumNumberOfHistories = maximumNumberOfHistories
    if walkStatus==ScolaWalker.TERMINATED:
      self._walkStarted = False
    return walkStatus

  def _WalkNext(self):
    if self._numberOfHistories>=self._maximumNumberOfHistories \
      or self._numberOfSteps>=self._maximumNumberOfStepsOfWalks \
      or self._walkStack.GetSize()>self._maximumSizeOfHistories:
      return
    self._numberOfSteps += 1

    if self._status==ScolaWalker.FORTH:
      choiceQueue = self._walkStack.GetTopChoiceQueue()
      (processNumber, branchPath) = choiceQueue.DequeueChoice()
      ok = self.MoveProcessForward(processNumber, branchPath)
      if ok:
        choiceQueue = self._walkStack.PushChoiceQueue()
        self._EnqueueChoices(choiceQueue)
        if choiceQueue.IsEmpty():
          self._numberOfHistories += 1
          # self._PrintWalkHistory()
          self._walkStack.PopChoiceQueue()
          self._status = ScolaWalker.BACK
      elif choiceQueue.IsEmpty():
        self._walkStack.PopChoiceQueue()
        self._status = ScolaWalker.BACK
      return ScolaWalker.CONTINUING

    if self._status==ScolaWalker.BACK:
      if self._walkStack.IsEmpty():
        return ScolaWalker.TERMINATED
      self._stepper.Backtrack()
      walkChoiceQueue = self._walkStack.GetTopChoiceQueue()
      if walkChoiceQueue.IsEmpty():
        self._walkStack.PopChoiceQueue()
        return ScolaWalker.CONTINUING
      self._status = ScolaWalker.FORTH
      return ScolaWalker.CONTINUING

    return ScolaWalker.TERMINATED # should not happen

  def _EnqueueChoices(self, choiceQueue):
    numbersOfActiveProcesses = self.GetNumbersOfActiveProcesses()
    for processNumber in numbersOfActiveProcesses:
      if self.IsProcessLocatedOnATerminalState(processNumber):
        continue
      if self.IsProcessLocatedOnAChoiceGateway(processNumber):
        paths = self.GetBranchesOfLocationOfProcess(processNumber)
        for path in paths:
          choiceQueue.EnqueueChoice(processNumber, path)
        continue
      choiceQueue.EnqueueChoice(processNumber, None)

  def _PrintWalkHistory(self):
    self.Println("#---")
    self.Println("#history " + str(self._numberOfHistories))
    self.PrintHistory()

# 3) WalkStack
# ------------

# 3.1) WalkChoiceQueue
# --------------------

class ScolaWalkChoiceQueue:
  def __init__(self):
    self._choices = []

  def IsEmpty(self):
    return len(self._choices)==0

  def DequeueChoice(self):
    return self._choices.pop(0)

  def EnqueueChoice(self, processNumber, branchPath):
    return self._choices.append((processNumber, branchPath))


# 3.2) ScolaWalkStack
# -------------------

class ScolaWalkStack:
  def __init__(self):
    self._choiceQueues = []

  def Reset(self):
    self._choiceQueues = []

  def IsEmpty(self):
    return len(self._choiceQueues)==0

  def GetSize(self):
    return len(self._choiceQueues)

  def PushChoiceQueue(self):
    choiceQueue = ScolaWalkChoiceQueue()
    self._choiceQueues.append(choiceQueue)
    return choiceQueue

  def PopChoiceQueue(self):
    return self._choiceQueues.pop()

  def GetTopChoiceQueue(self):
    return self._choiceQueues[-1]

  def Dump(self, status):
    traceFile = open("trace.txt", "a")
    traceFile.write("++ " + status +"\n")
    for choiceQueue in self._choiceQueues:
      for (processNumber, _) in choiceQueue._choices:
        traceFile.write("\t" + str(processNumber))
      traceFile.write("\n")
    traceFile.write("-- " + status +"\n")
    traceFile.flush()
    traceFile.close()