# Scola version 1.1.0
# ScolaExpression.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Expressions

# 1) Imported modules
# -------------------

import string
import sys

from ScolaMetaModel import *
from ScolaUniverse import *

# 2) Expressions
# --------------

class ScolaExpression:
  REFERENCE = "REFERENCE"
  SYMBOL = "SYMBOL"
  NONE = "NONE"
  # Boolean
  TRUE = "TRUE"
  FALSE = "FALSE"
  AND = "AND"
  OR = "OR"
  NOT = "NOT"
  # Inequalities
  EQ = "EQ"
  DF = "DF"
  LT = "LT"
  GT = "GT"
  LEQ = "LEQ"
  GEQ = "GET"
  # Arithmetic
  INTEGER = "INTEGER"
  REAL = "REAL"
  OPP = "OPP"
  INV = "INV"
  ADD = "ADD"
  SUB = "SUB"
  MUL = "MUL"
  DIV = "DIV"
  MIN = "MIN"
  MAX = "MAX"
  ABS = "ABS"
  POW = "POW"
  EXP = "EXP"
  LOG = "LOG"
  SQRT = "SQRT"
  MOD = "MOD"
  CEIL = "CEIL"
  FLOOR = "FLOOR"
  COUNT = "COUNT"
  CAST_INTEGER = "CAST_INTEGER"
  CAST_REAL = "CAST_REAL"
  # symbols
  CAST_SYMBOL = "CAST_SYMBOL",
  IDENTIFIER = "IDENTIFIER",
  OWNER = "OWNER",
  # strings
  STRING = "STRING"
  APPEND = "APPEND"
  PREPEND = "PREPEND"
  CAST_STRING = "CAST_STRING"
  IF = "IF"
  # paths
  IS_BLOCK = "IS_BLOCK"
  IS_PORT = "IS_PORT"
  IS_ASSERTION = "IS_ASSERTION"
  SIZE = "SIZE"
  ELEMENT = "ELEMENT"
  # Quote/Eval
  QUOTE = "QUOTE",
  EVAL = "EVAL"

  def __init__(self, type):
    self._type = type
    self._children = []
    self._parent = None
    self._value = ScolaConstant.NewNone()

  @staticmethod
  def NewReference(reference):
    self = ScolaExpression(ScolaExpression.REFERENCE)
    self._value = ScolaConstant.NewSymbol(reference)
    return self

  @staticmethod
  def NewSymbol(value):
    self = ScolaExpression(ScolaExpression.SYMBOL)
    self._value = ScolaConstant.NewSymbol(value)
    return self

  @staticmethod
  def NewNone():
    self = ScolaExpression(ScolaExpression.NONE)
    self._value = ScolaConstant.NewNone()
    return self

  @staticmethod
  def NewTrue():
    self = ScolaExpression(ScolaExpression.TRUE)
    self._value = ScolaConstant.NewTrue()
    return self

  @staticmethod
  def NewFalse():
    self = ScolaExpression(ScolaExpression.FALSE)
    self._value = ScolaConstant.NewFalse()
    return self

  @staticmethod
  def NewInteger(value):
    self = ScolaExpression(ScolaExpression.INTEGER)
    self._value = ScolaConstant.NewInteger(value)
    return self

  @staticmethod
  def NewReal(value):
    self = ScolaExpression(ScolaExpression.REAL)
    self._value = ScolaConstant.NewReal(value)
    return self

  @staticmethod
  def NewString(value):
    self = ScolaExpression(ScolaExpression.STRING)
    self._value = ScolaConstant.NewString(value)
    return self

  def GetType(self):
    return self._type

  def GetNumberOfChildren(self):
    return len(self._children)
  def AddChild(self, child):
    self._children.append(child)
    child._parent = self
  def AddChildren(self, children):
    for child in children:
      self.AddChild(child)
  def RemoveChild(self, child):
    self._children.remove(child)
    child._parent = None
  def RemoveChildren(self, children):
    for child in children:
      self.RemoveChild(child)
  def RemoveAllChildren(self):
    while self.GetNumberOfChildren()>0:
      child = self.GetChild(0)
      self.RemoveChild(child)
  def GetChild(self, index):
    if index<0 or index>=self.GetNumberOfChildren():
      return None
    return self._children[index]
  def SetChild(self, index, child):
    if index<0 or index>self.GetNumberOfChildren():
      raise ScolaInternalError("index out of the range")
    elif index==self.GetNumberOfChildren():
      self.AddChild(child)
    else:
      oldChild = self._children[index]
      oldChild._parent = None
      self._children[index] = child
      child._parent = self
  def GetChildren(self):
    return self._children

  def GetParent(self):
    return self._parent
  def SetParent(self, parent):
    if self._parent!=None:
      self._parent.RemoveChild(self)
    self._parent.AddChild(self)

  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value = value

  def DeepCopy(self):
    if self._type==ScolaExpression.NONE:
      return ScolaExpression.NewNone()
    if self._type==ScolaExpression.REFERENCE:
      return ScolaExpression.NewReference(self._value.GetValueAsSymbol())
    if self._type==ScolaExpression.SYMBOL:
      return ScolaExpression.NewSymbol(self._value.GetValueAsSymbol())
    if self._type==ScolaExpression.TRUE:
      return ScolaExpression.NewTrue()
    if self._type==ScolaExpression.FALSE:
      return ScolaExpression.NewFalse()
    if self._type==ScolaExpression.INTEGER:
      return ScolaExpression.NewInteger(self._value.GetValueAsInteger())
    if self._type==ScolaExpression.REAL:
      return ScolaExpression.NewReal(self._value.GetValueAsReal())
    if self._type==ScolaExpression.STRING:
      return ScolaExpression.NewString(self._value.GetValueAsString())
    copy = ScolaExpression(self._type)
    for child in self.GetChildren():
      childCopy = child.DeepCopy()
      copy.AddChild(childCopy)
    return copy

