# Scola version 1.1.0
# ScolaPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Printer

# 1) Imported modules
# -------------------

import string
import sys
import re
from ScolaLexer import *
from ScolaModels import *
from ScolaHistories import *
from ScolaProcesses import *
from ScolaBacktrackStack import *

# 2) Printer
# ----------

# 2.1) Creator
# ------------

class ScolaPrinter:
  _Keywords = dict()

  def __init__(self):
    self._depth = 0
    self._indentation = "  "
    self._outputFile = sys.stdout
    if len(ScolaPrinter._Keywords)==0:
      for keyword in ScolaLexer._Keywords:
        code = ScolaLexer._Keywords[keyword]
        ScolaPrinter._Keywords[code] = keyword

# 2.2) Models
# -----------

  def ExportModel(self, model, fileName):
    try:
      outputFile = open(fileName, "r")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    previousOutputFile = self._outputFile
    self._outputFile = outputFile
    self.PrintModel(model)
    self._outputFile = previousOutputFile
    outputFile.close()

  def PrintModel(self, model):
    self._depth = 0
    self._PrintElements(model)

  def _PrintContainer(self, container, keyword):
    self.PrintIndentation()
    self.PrintKeyword(keyword)
    if container.GetIdentifier()!=None:
      self.Print(" ")
      self.Print(container.GetIdentifier())
    elif container.GetPath()!=None:
      self.Print(" ")
      self.PrintPath(container.GetPath())
    if container.GetNumberOfElements()==0:
      self.Print(" ")
    else:
      self.Println()
      self._depth += 1
      self._PrintElements(container)
      self._depth -= 1
      self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def _PrintElements(self, container):
    first = True
    for element in container.GetElements():
      if first:
        first = False
      elif container.GetType()==ScolaModel.MODEL:
        self.Println()
      self._PrintElement(element)

  def _PrintElement(self, element):
    if element.GetType()==ScolaModel.MODEL:
      self.PrintModel(element)
    elif element.GetType()==ScolaModel.CLASS:
      self.PrintClass(element)
    elif element.GetType()==ScolaModel.INSTANCE:
      self.PrintInstance(element)
    elif element.GetType()==ScolaModel.CLONES:
      self.PrintClonesDirective(element)
    elif element.GetType()==ScolaModel.INHERITS:
      self.PrintInheritsDirective(element)
    elif element.GetType()==ScolaModel.DOMAIN:
      self.PrintDomain(element)
    elif element.GetType()==ScolaModel.BLOCK:
      self.PrintBlock(element)
    elif element.GetType()==ScolaModel.ASSERTION:
      self.PrintAssertion(element)
    elif element.GetType()==ScolaModel.PORT:
      self.PrintPort(element)
    elif element.GetType()==ScolaModel.SCENARIO:
      self.PrintScenario(element)
    elif element.GetType()==ScolaModel.STATE:
      self.PrintState(element)
    elif element.GetType()==ScolaModel.TASK:
      self.PrintTask(element)
    elif element.GetType()==ScolaModel.TEST:
      self.PrintTestGateway(element)
    elif element.GetType()==ScolaModel.CASE:
      self.PrintGatewayCase(element)
    elif element.GetType()==ScolaModel.CHOICE:
      self.PrintChoiceGateway(element)
    elif element.GetType()==ScolaModel.BRANCH:
      self.PrintGatewayBranch(element)
    elif element.GetType()==ScolaModel.FORK:
      self.PrintForkGateway(element)
    elif element.GetType()==ScolaModel.JOIN:
      self.PrintJoinGateway(element)
    elif element.GetType()==ScolaModel.SPLIT:
      self.PrintSplitGateway(element)
    elif element.GetType()==ScolaModel.MERGE:
      self.PrintMergeGateway(element)
    elif element.GetType()==ScolaModel.MEET:
      self.PrintMeetGateway(element)
    elif element.GetType()==ScolaModel.NEXT:
      self.PrintNextDirective(element)
    elif element.GetType()==ScolaModel.FAIL:
      self.PrintFailInstruction(element)
    elif element.GetType()==ScolaModel.SET:
      self.PrintSetInstruction(element)
    elif element.GetType()==ScolaModel.NEW_PORT:
      self.PrintNewPortInstruction(element)
    elif element.GetType()==ScolaModel.NEW_BLOCK:
      self.PrintNewBlockInstruction(element)
    elif element.GetType()==ScolaModel.DELETE:
      self.PrintDeleteInstruction(element)
    elif element.GetType()==ScolaModel.MOVE:
      self.PrintMoveInstruction(element)
    elif element.GetType()==ScolaModel.SEND:
      self.PrintSendInstruction(element)
    elif element.GetType()==ScolaModel.RECEIVE:
      self.PrintReceiveInstruction(element)
    elif element.GetType()==ScolaModel.CLONE:
      self.PrintCloneInstruction(element)
    elif element.GetType()==ScolaModel.IF:
      self.PrintIfInstruction(element)
    elif element.GetType()==ScolaModel.BEGIN:
      self.PrintInstructionBlock(element)
    elif element.GetType()==ScolaModel.PRINT:
      self.PrintPrintInstruction(element)

  def PrintClass(self, declaration):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_CLASS)
    self.Print(" ")
    self.Print(declaration.GetIdentifier())
    self.Print("(")
    identifier = declaration.GetIdentifierOfInheritedClass()
    if identifier!=None:
      self.Print(identifier)
    elif declaration.GetEventualType()==ScolaModel.BLOCK:
      self.PrintKeyword(ScolaToken.KEYWORD_BLOCK)
    elif declaration.GetEventualType()==ScolaModel.SCENARIO:
      self.PrintKeyword(ScolaToken.KEYWORD_SCENARIO)
    self.Print(")")
    if declaration.GetNumberOfElements()==0:
      self.Print(" ")
    else:
      self.Println()
      self._depth += 1
      self._PrintElements(declaration)
      self._depth -= 1
      self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintInstance(self, instance):
    self.PrintIndentation()
    self.Print(instance.GetClassIdentifier())
    self.Print(" ")
    self.Print(instance.GetIdentifier())
    if instance.GetAsPath()!=None:
      self.Print(" ")
      self.PrintKeyword(ScolaToken.KEYWORD_AS)
      self.Print(" ")
      self.PrintPath(instance.GetAsPath())
    if instance.GetNumberOfElements()==0:
      self.Print(" ")
    else:
      self.Println()
      self._depth += 1
      self._PrintElements(instance)
      self._depth -= 1
      self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintClonesDirective(self, clonesDirective):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_CLONE)
    self.Print(" ")
    self.PrintPath(clonesDirective.GetPathToClonedElement())
    self.Print(" ")
    self.PrintKeyword(ScolaToken.KEYWORD_AS)
    self.Print(" ")
    self.Print(clonesDirective.GetIdentifier())
    if clonesDirective.GetAsPath()!=None:
      self.Print(" ")
      self.PrintKeyword(ScolaToken.KEYWORD_AS)
      self.Print(" ")
      self.PrintPath(clonesDirective.GetAsPath())
    if clonesDirective.GetNumberOfElements()==0:
      self.Print(" ")
    else:
      self.Println()
      self._depth += 1
      self._PrintElements(clonesDirective)
      self._depth -= 1
      self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintInheritsDirective(self, directive):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_INHERITS)
    self.Print(" ")
    pathToInheritedElement = directive.GetPathToInheritedElement()
    if pathToInheritedElement!=None:
      self.PrintPath(pathToInheritedElement)
    elif directive.GetEventualType()==ScolaModel.BLOCK:
      self.PrintKeyword(ScolaToken.KEYWORD_BLOCK)
    elif directive.GetEventualType()==ScolaModel.SCENARIO:
      self.PrintKeyword(ScolaToken.KEYWORD_SCENARIO)
    self.Println()

  def PrintBaseType(self, baseType):
    if baseType.GetType()==ScolaBaseType.NONE:
      self.PrintKeyword(ScolaToken.KEYWORD_NONE)
    elif baseType.GetType()==ScolaBaseType.BOOLEAN:
      self.PrintKeyword(ScolaToken.KEYWORD_BOOLEAN)
    elif baseType.GetType()==ScolaBaseType.INTEGER:
      self.PrintKeyword(ScolaToken.KEYWORD_INTEGER)
    elif baseType.GetType()==ScolaBaseType.REAL:
      self.PrintKeyword(ScolaToken.KEYWORD_REAL)
    elif baseType.GetType()==ScolaBaseType.SYMBOL:
      self.PrintKeyword(ScolaToken.KEYWORD_SYMBOL)
    elif baseType.GetType()==ScolaBaseType.STRING:
      self.PrintKeyword(ScolaToken.KEYWORD_STRING)
    elif baseType.GetType()==ScolaBaseType.ANY:
      self.PrintKeyword(ScolaToken.KEYWORD_PORT)
    else:
     self.Print(baseType.GetIdentifier())

  def PrintDomain(self, domain):
    enumeration = domain.GetDomain()
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_DOMAIN)
    self.Print(" ")
    self.Print(enumeration.GetIdentifier())
    self.Print(" {")
    first = True
    for member in enumeration.GetMembers():
      if first:
        first = False
      else:
        self.Print(", ")
      self.PrintConstant(member)
    self.Print("}")
    semiLattice = enumeration.GetSemiLattice()
    if semiLattice==None:
      self.Print(" ")
      self.PrintKeyword(ScolaToken.KEYWORD_END)
      self.Println()
      return
    self.Println()
    self._depth += 1
    for constraint in semiLattice.GetOrderConstraints():
      self.PrintIndentation()
      self.PrintKeyword(ScolaToken.KEYWORD_LT)
      self.Print(" ")
      self.PrintConstant(constraint.GetMember1())
      self.Print(" ")
      self.PrintConstant(constraint.GetMember2())
      self.Println()
    self._depth -= 1
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintBlock(self, block):
    self._PrintContainer(block, ScolaToken.KEYWORD_BLOCK)

  def PrintPort(self, port):
    self.PrintIndentation()
    self.PrintBaseType(port.GetBaseType())
    self.Print(" ")
    self.Print(port.GetIdentifier())
    self.Print(" ")
    self.PrintExpression(port.GetInitialization())
    self.Println()

  def PrintAssertion(self, assertion):
    self._PrintContainer(assertion, ScolaToken.KEYWORD_ASSERTION)

  def PrintScenario(self, scenario):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_SCENARIO)
    self.Print(" ")
    self.Print(scenario.GetIdentifier())
    if scenario.GetAsPath()!=None:
      self.Print(" ")
      self.PrintKeyword(ScolaToken.KEYWORD_AS)
      self.Print(" ")
      self.PrintPath(scenario.GetAsPath())
    if scenario.GetNumberOfElements()==0:
      self.Print(" ")
    else:
      self.Println()
      self._depth += 1
      self._PrintElements(scenario)
      self._depth -= 1
      self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintState(self, state):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_STATE)
    self.Print(" ")
    self.Print(state.GetIdentifier())
    self.Println()

  def PrintTask(self, task):
    self._PrintContainer(task, ScolaToken.KEYWORD_TASK)

  def PrintTestGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_TEST)

  def PrintGatewayCase(self, gatewayCase):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_CASE)
    self.Print(" ")
    self.Print(gatewayCase.GetIdentifier())
    self.Print(" ")
    self.PrintExpression(gatewayCase.GetCondition())
    self.Println()

  def PrintChoiceGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_CHOICE)

  def PrintGatewayBranch(self, gatewayBranch):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_BRANCH)
    self.Print(" ")
    self.Print(gatewayBranch.GetIdentifier())
    self.Println()

  def PrintForkGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_FORK)

  def PrintJoinGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_JOIN)

  def PrintSplitGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_SPLIT)

  def PrintMergeGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_MERGE)

  def PrintMeetGateway(self, gateway):
    self._PrintContainer(gateway, ScolaToken.KEYWORD_MEET)

  def PrintNextDirective(self, connection):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_NEXT)
    self.Print(" ")
    self.PrintPath(connection.GetSourcePath())
    self.Print(" ")
    self.PrintPath(connection.GetTargetPath())
    self.Println()

  def PrintFailInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_FAIL)
    self.Println()

  def PrintSetInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_SET)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetValueExpression())
    self.Println()

  def PrintNewPortInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_NEW)
    self.Print(" ")
    self.PrintBaseType(instruction.GetBaseType())
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetInitializationExpression())
    self.Println()

  def PrintNewBlockInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_NEW)
    self.Print(" ")
    blockClass = instruction.GetBlockClass()
    if blockClass==None:
      self.PrintKeyword(ScolaToken.KEYWORD_BLOCK)
    else:
      self.Print(blockClass.GetIdentifier())
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Println()

  def PrintDeleteInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_DELETE)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Println()

  def PrintMoveInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_MOVE)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetTargetPathExpression())
    self.Println()

  def PrintSendInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_SEND)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetReceiverPathExpression())
    self.Println()

  def PrintReceiveInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_RECEIVE)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetSenderPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetTargetPathExpression())
    self.Println()

  def PrintCloneInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_CLONE)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Print(" ")
    self.PrintExpression(instruction.GetTargetPathExpression())
    self.Println()

  def PrintIfInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_IF)
    self.Print(" ")
    self.PrintExpression(instruction.GetCondition())
    self.Print(" ")
    self.PrintKeyword(ScolaToken.KEYWORD_THEN)
    self.Println()
    self._depth += 1
    self._PrintElement(instruction.GetThenInstruction())
    self._depth -= 1
    if instruction.GetElseInstruction()!=None:
      self.PrintIndentation()
      self.PrintKeyword(ScolaToken.KEYWORD_ELSE)
      self.Println()
      self._depth += 1
      self._PrintElement(instruction.GetElseInstruction())
      self._depth -= 1

  def PrintInstructionBlock(self, instructionBlock):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_BEGIN)
    self.Println()
    self._depth += 1
    self._PrintElements(instructionBlock)
    self._depth -= 1
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_END)
    self.Println()

  def PrintPrintInstruction(self, instruction):
    self.PrintIndentation()
    self.PrintKeyword(ScolaToken.KEYWORD_PRINT)
    self.Print(" ")
    self.PrintExpression(instruction.GetPathExpression())
    self.Println()

  def PrintExpression(self, expression):
    if expression.GetType()==ScolaExpression.REFERENCE:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.SYMBOL:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.NONE:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.TRUE:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.FALSE:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.AND:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_AND)
    elif expression.GetType()==ScolaExpression.OR:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_OR)
    elif expression.GetType()==ScolaExpression.NOT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_NOT)
    elif expression.GetType()==ScolaExpression.EQ:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_EQ)
    elif expression.GetType()==ScolaExpression.DF:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_DF)
    elif expression.GetType()==ScolaExpression.LT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_LT)
    elif expression.GetType()==ScolaExpression.GT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_GT)
    elif expression.GetType()==ScolaExpression.LEQ:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_LEQ)
    elif expression.GetType()==ScolaExpression.GEQ:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_GEQ)
    elif expression.GetType()==ScolaExpression.INTEGER:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.REAL:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.ADD:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_ADD)
    elif expression.GetType()==ScolaExpression.SUB:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_SUB)
    elif expression.GetType()==ScolaExpression.MUL:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_MUL)
    elif expression.GetType()==ScolaExpression.DIV:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_DIV)
    elif expression.GetType()==ScolaExpression.MIN:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_MIN)
    elif expression.GetType()==ScolaExpression.MAX:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_MAX)
    elif expression.GetType()==ScolaExpression.ABS:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_ABS)
    elif expression.GetType()==ScolaExpression.POW:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_POW)
    elif expression.GetType()==ScolaExpression.EXP:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_EXP)
    elif expression.GetType()==ScolaExpression.LOG:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_LOG)
    elif expression.GetType()==ScolaExpression.SQRT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_SQRT)
    elif expression.GetType()==ScolaExpression.MOD:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_MOD)
    elif expression.GetType()==ScolaExpression.CEIL:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_CEIL)
    elif expression.GetType()==ScolaExpression.FLOOR:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_FLOOR)
    elif expression.GetType()==ScolaExpression.COUNT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_COUNT)
    elif expression.GetType()==ScolaExpression.STRING:
      self.PrintConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.APPEND:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_APPEND)
    elif expression.GetType()==ScolaExpression.IF:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_IF)
    elif expression.GetType()==ScolaExpression.CAST_INTEGER:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_INTEGER)
    elif expression.GetType()==ScolaExpression.CAST_REAL:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_REAL)
    elif expression.GetType()==ScolaExpression.CAST_STRING:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_STRING)
    elif expression.GetType()==ScolaExpression.CAST_SYMBOL:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_SYMBOL)
    elif expression.GetType()==ScolaExpression.IDENTIFIER:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_IDENTIFIER)
    elif expression.GetType()==ScolaExpression.OWNER:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_OWNER)
    elif expression.GetType()==ScolaExpression.IS_BLOCK:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_IS_BLOCK)
    elif expression.GetType()==ScolaExpression.IS_PORT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_IS_PORT)
    elif expression.GetType()==ScolaExpression.IS_ASSERTION:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_IS_ASSERTION)
    elif expression.GetType()==ScolaExpression.SIZE:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_SIZE)
    elif expression.GetType()==ScolaExpression.ELEMENT:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_ELEMENT)
    elif expression.GetType()==ScolaExpression.QUOTE:
      childExpression = expression.GetChild(0)
      if childExpression.GetType()==ScolaExpression.SYMBOL or childExpression.GetType()==ScolaExpression.REFERENCE:
        self.Print("'")
        self.PrintExpression(childExpression)
      else:
        self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_QUOTE)
    elif expression.GetType()==ScolaExpression.EVAL:
      self._PrintParenthesedExpression(expression, ScolaToken.KEYWORD_EVAL)

  def _PrintParenthesedExpression(self, expression, keyword):
    self.Print("(")
    self.PrintKeyword(keyword)
    for child in expression.GetChildren():
      self.Print(" ")
      self.PrintExpression(child)
    self.Print(")")

  def PrintConstant(self, constant):
    if constant.GetType()==ScolaConstant.NONE:
      self.PrintKeyword(ScolaToken.KEYWORD_NONE)
    elif constant.GetType()==ScolaConstant.TRUE:
      self.PrintKeyword(ScolaToken.KEYWORD_TRUE)
    elif constant.GetType()==ScolaConstant.FALSE:
      self.PrintKeyword(ScolaToken.KEYWORD_FALSE)
    elif constant.GetType()==ScolaConstant.INTEGER:
      self.Print(str(constant.GetValueAsInteger()))
    elif constant.GetType()==ScolaConstant.REAL:
      self.Print(str(constant.GetValueAsReal()))
    elif constant.GetType()==ScolaConstant.SYMBOL:
      self.Print(constant.GetValueAsSymbol())
    elif constant.GetType()==ScolaConstant.STRING:
      self.Print('"' + constant.GetValueAsString() + '"')

  def PrintPath(self, path):
    self.Print(path.ToString())

# 2.3) Simulations
# ----------------

  def PrintValuesOfPortsOfBlock(self, element):
    if element.GetType()==ScolaModel.BLOCK:
      self.PrintIndentation()
      self.PrintKeyword(ScolaToken.KEYWORD_BLOCK)
      self.Print(" ")
      self.Print(element.GetIdentifier())
      self.Println()
      for child in element.GetElements():
        self._depth += 1
        self.PrintValuesOfPortsOfBlock(child)
        self._depth -= 1
      # for identifier in element.GetDictonary():
        # self._depth += 1
        # self.PrintIndentation()
        # self.Print("-- ")
        # self.Print(identifier)
        # self.Println()
        # self._depth -= 1
      self.PrintIndentation()
      self.PrintKeyword(ScolaToken.KEYWORD_END)
      self.Println()
    elif element.GetType()==ScolaModel.PORT:
      self.PrintValueOfPort(element)

  def PrintValueOfPort(self, port):
    self.PrintIndentation()
    self.PrintBaseType(port.GetBaseType())
    self.Print(" ")
    self.Print(port.GetIdentifier())
    self.Print(" ")
    self.PrintConstant(port.GetValue())
    self.Println()

  def PrintProcess(self, process):
    self.PrintInteger(process.GetNumber())
    self.Print("\t")
    parent = process.GetParent()
    if parent==None:
      self.Print("_")
    else:
      self.PrintInteger(parent.GetNumber())
    self.Print("\t")
    if process.IsActive():
      element = process.GetLocation()
      self.PrintScenarioElement(element)
    else:
      self.Print("inactive")
    self.Println()

  def PrintScenarioElement(self, element):
    if element.GetType()==ScolaModel.STATE:
      self.PrintKeyword(ScolaToken.KEYWORD_STATE)
    elif element.GetType()==ScolaModel.TASK:
      self.PrintKeyword(ScolaToken.KEYWORD_TASK)
    elif element.GetType()==ScolaModel.TEST:
      self.PrintKeyword(ScolaToken.KEYWORD_TEST)
    elif element.GetType()==ScolaModel.CASE:
      self.PrintKeyword(ScolaToken.KEYWORD_CASE)
    elif element.GetType()==ScolaModel.FORK:
      self.PrintKeyword(ScolaToken.KEYWORD_FORK)
    elif element.GetType()==ScolaModel.CHOICE:
      self.PrintKeyword(ScolaToken.KEYWORD_CHOICE)
    elif element.GetType()==ScolaModel.BRANCH:
      self.PrintKeyword(ScolaToken.KEYWORD_BRANCH)
    elif element.GetType()==ScolaModel.JOIN:
      self.PrintKeyword(ScolaToken.KEYWORD_JOIN)
    elif element.GetType()==ScolaModel.MERGE:
      self.PrintKeyword(ScolaToken.KEYWORD_MERGE)
    self.Print("\t")
    path = element.GetAbsolutePath()
    self.PrintPath(path)

  def PrintNextLocationsOfProcess(self, process):
    self.Print("process ")
    self.PrintInteger(process.GetNumber())
    self.Print(":")
    self.Println()
    if not process.IsActive():
      self.Print("\tinactive")
      self.Println()
      return
    element = process.GetLocation()
    if element.GetType()==ScolaModel.STATE:
      nextElement = element.GetNextElement()
      if nextElement!=None:
        self.Print("\t")
        self.PrintScenarioElement(nextElement)
        self.Println()
    elif element.GetType()==ScolaModel.TASK:
      nextElement = element.GetNextElement()
      if nextElement!=None:
        self.Print("\t")
        self.PrintScenarioElement(nextElement)
        self.Println()
    elif element.GetType()==ScolaModel.TEST:
      for childElement in element.GetElements():
        nextElement = childElement.GetNextElement()
        if nextElement!=None:
          self.Print("\t")
          self.PrintScenarioElement(childElement)
          self.Print("\t")
          self.PrintScenarioElement(nextElement)
          self.Println()
    elif element.GetType()==ScolaModel.FORK:
      for childElement in element.GetElements():
        nextElement = childElement.GetNextElement()
        if nextElement!=None:
          self.Print("\t")
          self.PrintScenarioElement(childElement)
          self.Print("\t")
          self.PrintScenarioElement(nextElement)
          self.Println()
    elif element.GetType()==ScolaModel.CHOICE:
      for childElement in element.GetElements():
        nextElement = childElement.GetNextElement()
        if nextElement!=None:
          self.Print("\t")
          self.PrintScenarioElement(childElement)
          self.Print("\t")
          self.Print(nextElement.GetIdentifier())
          self.Print("\t#")
          self.PrintScenarioElement(nextElement)
          self.Println()
    elif element.GetType()==ScolaModel.MERGE:
      nextElement = element.GetNextElement()
      if nextElement!=None:
        self.Print("\t")
        self.PrintScenarioElement(nextElement)
        self.Println()

  def PrintHistories(self, historyManager):
    for index in range(0, historyManager.GetNumberOfHistories()):
      self.Print("--- history ")
      self.PrintInteger(index)
      self.Println()
      history = historyManager.GetHistory(index)
      self.PrintHistory(history)

  def PrintHistory(self, history):
    for index in range(0, history.GetSize()):
      command = history.GetCommand(index)
      self.PrintCommand(command)

  def PrintCommand(self, command):
    if command.GetType()==ScolaCommand.NEW_PROCESS:
      self.Print("new ")
      self.Print(command.GetArgumentAsLocationName(0))
    elif command.GetType()==ScolaCommand.KILL_PROCESS:
      self.Print("kill ")
      self.PrintInteger(command.GetArgumentAsProcessNumber(0))
    elif command.GetType()==ScolaCommand.NEXT:
      self.Print("next ")
      self.PrintInteger(command.GetArgumentAsProcessNumber(0))
    elif command.GetType()==ScolaCommand.CHOICE:
      self.Print("next ")
      self.PrintInteger(command.GetArgumentAsProcessNumber(0))
      self.Print(" ")
      self.Print(command.GetArgumentAsLocationName(1))
    if command.GetComment()!=None:
      self.Print("\t#")
      self.Print(command.GetComment())
    self.Println()

# 2.4) Miscellaneous
# ------------------

  def GetOutputFile(self):
    return self._outputFile
  def SetOutputFile(self, outputFile):
    self._outputFile = outputFile

  def GetIndentation(self):
    return self._indentation
  def SetIndentation(self, indentation):
    self._indentation = indentation

  def PrintKeyword(self, keyword):
    self.Print(ScolaPrinter._Keywords[keyword])

  def PrintIndentation(self):
    for d in range(0, self._depth):
      self.Print(self._indentation)

  def PrintInteger(self, integer):
    self.Print(str(integer))

  def PrintReal(self, real):
    self.Print(str(real))

  def Print(self, *texts):
    for i in range(0, len(texts)):
      self._outputFile.write(texts[i])

  def Println(self, *texts):
    self.Print(*texts)
    self._outputFile.write("\n")

  def Flush(self):
    self._outputFile.flush()

# 2.5) Get strings
# ----------------

  def GetStringOfScenarioElement(self, element):
    if element.GetType()==ScolaModel.STATE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_STATE)
    elif element.GetType()==ScolaModel.TASK:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_TASK)
    elif element.GetType()==ScolaModel.TEST:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_TEST)
    elif element.GetType()==ScolaModel.CASE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_CASE)
    elif element.GetType()==ScolaModel.CHOICE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_CHOICE)
    elif element.GetType()==ScolaModel.BRANCH:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_BRANCH)
    elif element.GetType()==ScolaModel.FORK:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_FORK)
    elif element.GetType()==ScolaModel.JOIN:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_JOIN)
    elif element.GetType()==ScolaModel.SPLIT:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_SPLIT)
    elif element.GetType()==ScolaModel.MERGE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_MERGE)
    elif element.GetType()==ScolaModel.MEET:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_MEET)
    string += " "
    path = element.GetAbsolutePath()
    string += path.ToString()
    return string

  def GetStringOfExpression(self, expression):
    if expression.GetType()==ScolaExpression.REFERENCE:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.SYMBOL:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.NONE:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.TRUE:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.FALSE:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.AND:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_AND)
    elif expression.GetType()==ScolaExpression.OR:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_OR)
    elif expression.GetType()==ScolaExpression.NOT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_NOT)
    elif expression.GetType()==ScolaExpression.EQ:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_EQ)
    elif expression.GetType()==ScolaExpression.DF:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_DF)
    elif expression.GetType()==ScolaExpression.LT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_LT)
    elif expression.GetType()==ScolaExpression.GT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_GT)
    elif expression.GetType()==ScolaExpression.LEQ:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_LEQ)
    elif expression.GetType()==ScolaExpression.GEQ:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_GEQ)
    elif expression.GetType()==ScolaExpression.INTEGER:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.REAL:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.ADD:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_ADD)
    elif expression.GetType()==ScolaExpression.SUB:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_SUB)
    elif expression.GetType()==ScolaExpression.MUL:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_MUL)
    elif expression.GetType()==ScolaExpression.DIV:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_DIV)
    elif expression.GetType()==ScolaExpression.MIN:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_MIN)
    elif expression.GetType()==ScolaExpression.MAX:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_MAX)
    elif expression.GetType()==ScolaExpression.ABS:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_ABS)
    elif expression.GetType()==ScolaExpression.POW:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_POW)
    elif expression.GetType()==ScolaExpression.EXP:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_EXP)
    elif expression.GetType()==ScolaExpression.LOG:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_LOG)
    elif expression.GetType()==ScolaExpression.SQRT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_SQRT)
    elif expression.GetType()==ScolaExpression.MOD:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_MOD)
    elif expression.GetType()==ScolaExpression.CEIL:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_CEIL)
    elif expression.GetType()==ScolaExpression.FLOOR:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_FLOOR)
    elif expression.GetType()==ScolaExpression.COUNT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_COUNT)
    elif expression.GetType()==ScolaExpression.STRING:
      string = self.GetStringOfConstant(expression.GetValue())
    elif expression.GetType()==ScolaExpression.APPEND:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_APPEND)
    elif expression.GetType()==ScolaExpression.IF:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_IF)
    elif expression.GetType()==ScolaExpression.CAST_INTEGER:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_INTEGER)
    elif expression.GetType()==ScolaExpression.CAST_REAL:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_REAL)
    elif expression.GetType()==ScolaExpression.CAST_STRING:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_STRING)
    elif expression.GetType()==ScolaExpression.CAST_SYMBOL:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_SYMBOL)
    elif expression.GetType()==ScolaExpression.IDENTIFIER:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_IDENTIFIER)
    elif expression.GetType()==ScolaExpression.OWNER:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_OWNER)
    elif expression.GetType()==ScolaExpression.IS_PORT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_IS_PORT)
    elif expression.GetType()==ScolaExpression.IS_BLOCK:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_IS_BLOCK)
    elif expression.GetType()==ScolaExpression.IS_ASSERTION:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_IS_ASSERTION)
    elif expression.GetType()==ScolaExpression.SIZE:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_SIZE)
    elif expression.GetType()==ScolaExpression.ELEMENT:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_ELEMENT)
    elif expression.GetType()==ScolaExpression.QUOTE:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_QUOTE)
    elif expression.GetType()==ScolaExpression.EVAL:
      string = self.GetStringOfParenthesedExpression(expression, ScolaToken.KEYWORD_EVAL)
    else:
      string = "" # should not occur
    return string

  def GetStringOfParenthesedExpression(self, expression, keyword):
    string = "("
    string += self.GetStringOfKeyword(keyword)
    for child in expression.GetChildren():
      string += " "
      string += self.GetStringOfExpression(child)
    string += ")"
    return string

  def GetStringOfConstant(self, constant):
    if constant.GetType()==ScolaConstant.NONE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_NONE)
    elif constant.GetType()==ScolaConstant.TRUE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_TRUE)
    elif constant.GetType()==ScolaConstant.FALSE:
      string = self.GetStringOfKeyword(ScolaToken.KEYWORD_FALSE)
    elif constant.GetType()==ScolaConstant.INTEGER:
      string = str(constant.GetValueAsInteger())
    elif constant.GetType()==ScolaConstant.REAL:
      string = str(constant.GetValueAsReal())
    elif constant.GetType()==ScolaConstant.SYMBOL:
      string = constant.GetValueAsSymbol()
    elif constant.GetType()==ScolaConstant.STRING:
      string = '"' + constant.GetValueAsString() + '"'
    else:
      string = "" # should not occur
    return string

  def GetStringOfKeyword(self, keyword):
    return ScolaPrinter._Keywords[keyword]
